import PyIPSDK

import PyIPSDK.IPSDKIPLUtility as util
import PyIPSDK.IPSDKIPLGeometricTransform as gtrans
import PyIPSDK.IPSDKIPLGlobalMeasure as glbmsr
import PyIPSDK.IPSDKIPLArithmetic as arithm
import PyIPSDK.IPSDKIPLFiltering as filtering

import numpy as np

def registration(imageFixed,imageMoving,nbPointsRegistration):

	import PyIPSDK.IPSDKIPLRegistration as registration

	def getTransform(greySignaturesFixed, greySignaturesMoving, correlationThreshold, estimationConfig):

		# Calcul du recalage  appliquer en utilisant les descripteurs de chaque image

		# On commence par trier les paires de descripteurs, en gardant celles qui ont une translation infrieure au seuil [maxTx,maxTy] et une rotation suffisament faible (infrieure  thetaBin)

		thetaBin = 0.05
		maxTx = 10
		maxTy = 10

		outRegistrationResult = registration.greySignedFeatures2dRegistration(greySignaturesMoving, greySignaturesFixed, PyIPSDK.eRegistrationMotionModel2d.eRMM2d_Rigid, correlationThreshold, estimationConfig)

		dictPt = {}
		i = 0

		for ind in outRegistrationResult.indicators.assignments:
			dictPt[i] = {}
			dictPt[i]["x1"] = greySignaturesFixed.signatureColl[ind.ptIdx2].x
			dictPt[i]["y1"] = greySignaturesFixed.signatureColl[ind.ptIdx2].y
			dictPt[i]["x2"] = greySignaturesMoving.signatureColl[ind.ptIdx1].x
			dictPt[i]["y2"] = greySignaturesMoving.signatureColl[ind.ptIdx1].y
			dictPt[i]["corr"] = 0

			i += 1

		for i in dictPt:
			for j in dictPt:
				if j != i:
					v1x, v1y = dictPt[i]["x1"] - dictPt[j]["x1"], dictPt[i]["y1"] - dictPt[j]["y1"]
					v2x, v2y = dictPt[i]["x2"] - dictPt[j]["x2"], dictPt[i]["y2"] - dictPt[j]["y2"]

					value = (v1x * v2x + v1y * v2y) / (np.sqrt(v1x * v1x + v1y * v1y) * np.sqrt(v2x * v2x + v2y * v2y))
					value = max(-1, value)
					value = min(value, 1)

					signe = v2x * v1y - v2y * v1x
					if signe < 0:
						signe = -1
					else:
						signe = 1

					thetaValue = (np.arccos(value) * signe)
					if thetaValue >= -thetaBin and thetaValue <= thetaBin:
						dictPt[i]["corr"] += 1

		#On calcule ensuite la translation moyenne de l'ensemble des paires de descripteurs

		nbPoints = 0
		Tx = 0
		Ty = 0
		for i in dictPt:
			if dictPt[i]["corr"] >= 3 and abs(dictPt[i]["x2"] - dictPt[i]["x1"]) <= maxTx and abs(dictPt[i]["y2"] - dictPt[i]["y1"]) <= maxTy:
				Tx += (dictPt[i]["x2"] - dictPt[i]["x1"]) * dictPt[i]["corr"]
				Ty += (dictPt[i]["y2"] - dictPt[i]["y1"]) * dictPt[i]["corr"]
				nbPoints += dictPt[i]["corr"]

		if nbPoints != 0:
			Tx /= nbPoints
			Ty /= nbPoints

		return -Tx, -Ty, nbPoints

	def correctImage(image):

		# On augmente le contraste de l'image pour reprer plus efficacement les descripteurs

		image = filtering.median2dImg(image, 3, 3)

        statsMsr = glbmsr.statsMsr2d(image)

        histogramMsrParams = PyIPSDK.createHistoMsrParamsWithBinWidth(statsMsr.min, statsMsr.max, (statsMsr.max-statsMsr.min)/255)
		histogramMs2drResult = glbmsr.histogramMsr2d(image, histogramMsrParams)
		frequencies = histogramMs2drResult.frequencies
		sizeImage = image.getSizeX() * image.getSizeY()

		sumFreq = 0
		minFreq = None
		maxFreq = None
		for numFreq in range(len(frequencies)):

			sumFreq += frequencies[numFreq]
			if sumFreq >= sizeImage * 0.1 and minFreq is None:
				minFreq = numFreq
			if sumFreq >= sizeImage * 0.9 and maxFreq is None:
				maxFreq = numFreq

		maxFreq = max(maxFreq, minFreq + 1)

		image = arithm.addScalarImg(image, -minFreq)
		image = arithm.multiplyScalarImg(image, 255 / (maxFreq - minFreq))
		image = util.convertImg(image, PyIPSDK.eIBT_UInt8)

		return image
	
	#Application du recalage
	
	#Paramtres du recalage
	correlationThreshold = 0.95
	estimationConfig = PyIPSDK.EstimationConfig()
	estimationConfig.initLMS(0.48)
	
	#Correction du contraste et extraction des features
	correctImageFixed = correctImage(imageFixed)
	correctImageMoving = correctImage(imageMoving)
	greySignaturesFixed = registration.extractGreySignedFeatures2d(correctImageFixed, nbPointsRegistration)
	greySignaturesMoving = registration.extractGreySignedFeatures2d(correctImageMoving, nbPointsRegistration)
	
	#Calcul du vecteur de translation
	Tx, Ty, nbPoints = getTransform(greySignaturesFixed, greySignaturesMoving, correlationThreshold, estimationConfig)
	
	#Application de la translation
	transform1 = PyIPSDK.createWarpMotionTransform2d(PyIPSDK.eGeometricTransform2dType.eGT2DT_Translation,[Tx,Ty])
	outImage = PyIPSDK.createImage(imageMoving)
	util.eraseImg(outImage, 0)
	gtrans.warp2dImg(imageMoving, transform1, outImage)

	return outImage

